#include "Dict.h"

#include <iostream>

using namespace std;

namespace dictionary {

struct Dict::Node
{
    string key;
    unsigned value;
    Node* left;
    Node* right;

    Node* deep_clone() const;

    static void deep_delete(Node*);
    static Node* deep_clone(const Node*);
};

void Dict::Node::deep_delete(Node* node)
{
    if (node) {
        deep_delete(node->left);
        deep_delete(node->right);
        delete node;
    }
}

Dict::Node* Dict::Node::deep_clone(const Node* node)
{
    return node ? node->deep_clone() : nullptr;
}

Dict::Node* Dict::Node::deep_clone() const
{
    return new Node{key, value, deep_clone(left), deep_clone(right)};
}

Dict::Dict() : size_(0), root_(nullptr)
{ }

Dict::Dict(const Dict& that)
        : size_(that.size_),
          root_(Node::deep_clone(that.root_))
{ }

Dict& Dict::operator=(const Dict& that)
{
    Node::deep_delete(root_);
    size_ = that.size_;
    root_ = Node::deep_clone(that.root_);

    return *this;
}

Dict::~Dict()
{
    Node::deep_delete(root_);
}

void Dict::swap(Dict& other)
{
    std::swap(root_, other.root_);
    std::swap(size_, other.size_);
}

bool Dict::empty() const {
    return root_ == nullptr;
}

size_t Dict::size() const {
    return size_;
}

void Dict::clear()
{
    Node::deep_delete(root_);
    root_ = nullptr;
    size_ = 0;
}

bool Dict::member(const string& key) const
{
    return find_subtree_(root_, key) != nullptr;
}

unsigned Dict::lookup(const string& key) const
{
    Node* node = find_subtree_(root_, key);
    if (node == nullptr)
        throw Not_found(key);
    else return node->value;
}

unsigned& Dict::lookup(const std::string& key)
{
    Node* node = find_subtree_(root_, key);
    if (node == nullptr)
        throw Not_found(key);
    else return node->value;
}

void Dict::insert(const string& key, unsigned value)
{
    Node* & node = find_subtree_mut_(root_, key);
    if (node == nullptr) {
        node = new Node{key, value, nullptr, nullptr};
    } else {
        node->value = value;
    }
}

unsigned& Dict::operator[](const std::string& key)
{
    Node* & node = find_subtree_mut_(root_, key);
    if (node == nullptr) {
        node = new Node{key, 0, nullptr, nullptr};
        ++size_;
    }

    return node->value;
}

Dict::Node* Dict::find_subtree_(Dict::Node* link, const string& key)
{
    while (link != nullptr && link->key != key)
        link = (key < link->key) ? link->left : link->right;

    return link;
}

Dict::Node*& Dict::find_subtree_mut_(Dict::Node*& link, const string& key)
{
    Dict::Node** curr = &link;

    while (*curr) {
        if (key < (*curr)->key) {
            curr = &(*curr)->left;
        } else if (key > (*curr)->key) {
            curr = &(*curr)->right;
        } else break;
    }

    return *curr;
}

void Dict::get_keys_helper_(Dict::Node* node, vector<string>& out)
{
    if (node != nullptr) {
        get_keys_helper_(node->left, out);
        out.push_back(node->key);
        get_keys_helper_(node->right, out);
    }
}

vector<string> Dict::get_keys() const
{
    vector<string> result;
    get_keys_helper_(root_, result);
    return result;
}

void swap(Dict& d1, Dict& d2)
{
    d1.swap(d2);
}

}
